/*******************************TWI includes*************************/
#include <avr/io.h>
#include <avr/eeprom.h>
#include <compat/twi.h>
#include <avr/interrupt.h>
#include <string.h>

#define	VERSION		0x02

//MAX7313 TWI address
#define BLOCK0 0x10  //0001-0000
#define BLOCK1 0x20  //0010-0000
#define BLOCK2 0x28  //0010-1000
#define BLOCK3 0x50  //0101-0000
const unsigned char GROUPID[] = {0x00,0x01,0x02,0x03};

static volatile unsigned char twiunusable[]={0,0};
static volatile unsigned char twierror[]={0,0};

//Errors
#define ERRCOMMAND	0
#define ERRARGS		1
#define ERRTWI 		2

//Main Functions
void calibosc(void);
void pinconfig(void);
static __inline__ void wait(uint16_t);

//Chip Operation
unsigned char blockinit(void);
unsigned char blockwrite(unsigned char*);
unsigned char groupinit(unsigned char);
unsigned char groupwrite(unsigned char, unsigned char*);
unsigned char driverinit(unsigned char);
unsigned char driverwrite(unsigned char, unsigned char*);

//TWI Status
void twiseterror(unsigned char);
void twisetunusable(unsigned char);
void twisetgood(unsigned char);
unsigned char twicheckunusable(unsigned char);
unsigned char twilookupindex(unsigned char);

//TWI Send
void twiinit(void);
unsigned char twistart(unsigned char); 
void twistop(void);
unsigned char twiwrite(unsigned char);
unsigned char twiread(unsigned char);
unsigned char twitransmit(unsigned char, unsigned char[], unsigned char);
unsigned char twireceive(unsigned char,unsigned char*, unsigned char);

int main(void)
{
	calibosc();
	pinconfig();
	blockinit();
	return 0;
}

/******************************Main Functions************************/
void calibosc(void)
{
	unsigned char temp;
	temp=eeprom_read_byte((unsigned char*)1);
	if (temp>0x80 && temp<0xAF)
	{
		OSCCAL=(unsigned char)(1.036*(float)temp);
	}
}

void pinconfig(void)
{
	DDRD=0x0A;
	DDRC=0x03;
	PORTD=0xF7;
	PORTC=0x00;
	twiinit();
}

static __inline__ void wait(uint16_t count)
{
	__asm__ volatile
	(
		"1: sbiw %0,1" "\n\t"
		"brne 1b"
		: "=w" (count)
		: "0" (count)
	);
}

/*******************************Chip Operation***********************/
unsigned char blockinit(void)
{
	unsigned char i,err=0;
	for (i=0;i<4;i++)
	{
			err|=groupinit(i);
	}
	return err;
}

unsigned char blockwrite(unsigned char *buffer)
{
	unsigned char err;
	err=groupwrite(0,buffer);
	err|=groupwrite(1,buffer+24);
	err|=groupwrite(2,buffer+48);
	err|=groupwrite(3,buffer+72);
	return err;
}

unsigned char groupinit(unsigned char row)
{
	unsigned char err;
	err=driverinit(BLOCK0|GROUPID[row]);
	err|=driverinit(BLOCK1|GROUPID[row]);
	err|=driverinit(BLOCK2|GROUPID[row]);
	err|=driverinit(BLOCK3|GROUPID[row]);
	return err;
}

unsigned char groupwrite(unsigned char row, unsigned char *buffer)
{
	unsigned char err;
	err=driverwrite(BLOCK0|GROUPID[row],buffer);
	err|=driverwrite(BLOCK1|GROUPID[row],buffer+6);
	err|=driverwrite(BLOCK2|GROUPID[row],buffer+12);
	err|=driverwrite(BLOCK3|GROUPID[row],buffer+18);
	return err;
}

unsigned char driverinit(unsigned char address)
{
	unsigned char message[3];
	message[0] = 0x0F;
	message[1] = 0x08;
	if (twitransmit(address, message, 2 ))
	{
		twisetunusable(address);
		return 1<<ERRTWI;
	}
	message[0] = 0x06; 
	message[1] = 0x00;
	message[2] = 0x00;
	if (twitransmit(address, message, 3 ))
	{
		twisetunusable(address);
		return 1<<ERRTWI;
	}
	message[0] = 0x02; 
	message[1] = 0xFF;
	message[2] = 0xFF;
	if (twitransmit(address, message, 3 ))
	{
		twisetunusable(address);
		return 1<<ERRTWI;
	}
	twisetgood(address);
	return 0;
}

unsigned char driverwrite(unsigned char address, unsigned char *buffer)
{
	unsigned char i;
	unsigned char message[9];
	if (twicheckunusable(address))
	{
		twiseterror(address);
		return 1<<ERRTWI;
	}
	message[0]=0x10;
	for (i=0;i<8;i++)
		message[i+1]=buffer[i];
	if (twitransmit(address,message,9))
	{
		twiseterror(address);
		return 1<<ERRTWI;
	}
	twisetgood(address);
	return 0;
}

/*******************************TWI Status***************************/
void twiseterror(unsigned char address)
{
	unsigned char twiadd;
	twiadd=twilookupindex(address);
	if (twiadd<16)
	{
		twierror[twiadd>>3]|=(1<<(twiadd&0b00000111));
	}
}

void twisetunusable(unsigned char address)
{	
	unsigned char twiadd;
	twiadd=twilookupindex(address);
	if (twiadd<16)
	{
		twiunusable[twiadd>>3]|=(1<<(twiadd&0b00000111));
		twierror[twiadd>>3]|=(1<<(twiadd&0b00000111));
	}
}

void twisetgood(unsigned char address)
{
	unsigned char twiadd;
	twiadd=twilookupindex(address);
	if (twiadd<16)
	{
		twiunusable[twiadd>>3]&=~(1<<(twiadd&0b00000111));
		twierror[twiadd>>3]&=~(1<<(twiadd&0b00000111));
	}
}

unsigned char twicheckunusable(unsigned char address)
{
	unsigned char twiadd;
	twiadd=twilookupindex(address);
	if (twiadd==0xFF || (twiunusable[twiadd>>3]&(1<<(twiadd&0b00000111))))
	{
		return 1;
	}
	return 0;
}

unsigned char twilookupindex(unsigned char address)
{
	unsigned char i;
	for (i=0;i<4;i++)
	{
		if(address==(BLOCK0|GROUPID[i]))
		{
			return 0|i<<2;
		}
		if(address==(BLOCK1|GROUPID[i]))
		{
			return 1|i<<2;
		}
		if(address==(BLOCK2|GROUPID[i]))
		{
			return 2|i<<2;
		}
		if(address==(BLOCK3|GROUPID[i]))
		{
			return 3|i<<2;
		}
	}
	return 0xFF;
}

/*******************************TWI Send*****************************/
void twiinit(void)
{
	TWSR = 0;
	TWBR =0x0A;
}

unsigned char twistart(unsigned char address) 
{
	TWCR = (1<<TWEN) | (1<<TWSTA) | (1<<TWINT);
	while(!(TWCR & (1<<TWINT)))
	{
		asm volatile("nop"::);
	}
	if ( ((TW_STATUS & 0xF8) != TW_START) && ((TW_STATUS & 0xF8) != TW_REP_START)) 
	{
		return 1;
	}
	TWDR = address;
	TWCR = (1<<TWINT) | (1<<TWEN);
	while(!(TWCR & (1<<TWINT)))
	{
		asm volatile("nop"::);
	}
	if(((TW_STATUS & 0xF8) != TW_MT_SLA_ACK) && ((TW_STATUS & 0xF8) != TW_MR_SLA_ACK))
	{
		return 1;
	}
	return 0;
}

void twistop(void)
{
	TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWSTO);
	while(TWCR & (1<<TWSTO))
	{
		asm volatile("nop"::);
	}
}

unsigned char twiwrite(unsigned char data)
{
	TWDR = data;
	TWCR = (1<<TWINT) | (1<<TWEN);
	while(!(TWCR & (1<<TWINT)))
	{
		asm volatile("nop"::);
	}
	if( (TW_STATUS & 0xF8) != TW_MT_DATA_ACK)
	{
		return 1;
	}
	return 0;
}

unsigned char twiread(unsigned char ack)
{
	TWCR = (1<<TWINT) | (1<<TWEN) | (ack<<TWEA);
	while(!(TWCR & (1<<TWINT)))
	{
		asm volatile("nop"::);
	}
    return TWDR;
}

unsigned char twitransmit(unsigned char address, unsigned char msg[], unsigned char size)
{
	unsigned char n,error;
	error=twistart(address<<1);
	for (n=0;n<size && !error;n++)
		{
			error|=twiwrite(msg[n]);
    	}
	twistop();
	return error;
}

unsigned char twireceive(unsigned char address,unsigned char *buffer, unsigned char size)
{
	unsigned char n,error;
	error=twistart((address<<1)|1);
	for (n=size;n>1;n--) 
	{
		*buffer=twiread(1);
		buffer++;
	}
	*buffer=twiread(0);
	twistop();
	return error;
}
